@@ -41,7 +41,7 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  *
- * msnd_pinnacle.c,v 1.12 2005/01/31 16:07:20 cladisch Exp
+ * $Id: msnd_pinnacle.c,v 1.8 2000/12/30 00:33:21 sycamore Exp $
  *
  * 12-3-2000  Modified IO port validation  Steve Sycamore
  *
@@ -67,6 +67,7 @@
 #include <linux/smp_lock.h>
 #include <linux/vmalloc.h>
 #include <linux/ioport.h>
+#include <linux/firmware.h>
 #include <asm/irq.h>
 #include <asm/io.h>
 
@@ -91,13 +92,11 @@
 #  define LOGNAME			"snd_msnd_pinnacle"
 #endif
 
-extern int mod_firmware_load(const char *fn, char **fp);
-
 
 #define chip_t snd_msndpinnacle_pcm_t
 
 typedef struct snd_msndpinnacle_pcm {
-	snd_card_t*		card;
+	struct snd_card *		card;
 	spinlock_t lock;
 	struct timer_list timer;
 	unsigned int pcm_size;
@@ -107,16 +106,16 @@ typedef struct snd_msndpinnacle_pcm {
 	unsigned int pcm_jiffie;	/* bytes per one jiffie */
 	unsigned int pcm_irq_pos;	/* IRQ position */
 	unsigned int pcm_buf_pos;	/* position in buffer */
-	snd_pcm_substream_t *substream;
+	struct snd_pcm_substream *substream;
 } snd_msndpinnacle_pcm_t;
 
 
 static int index[SNDRV_CARDS] = SNDRV_DEFAULT_IDX;	/* Index 0-MAX */
 static char *id[SNDRV_CARDS] = SNDRV_DEFAULT_STR;	/* ID for this card */
 
-MODULE_PARM(index, "1-" __MODULE_STRING(SNDRV_CARDS) "i");
+module_param_array(index, int, NULL, S_IRUGO);
 MODULE_PARM_DESC(index, "Index value for msnd_pinnacle soundcard.");
-MODULE_PARM(id, "1-" __MODULE_STRING(SNDRV_CARDS) "s");
+module_param_array(id, charp, NULL, S_IRUGO);
 MODULE_PARM_DESC(id, "ID string for msnd_pinnacle soundcard.");
 
 #ifndef CONFIG_MSND_WRITE_NDELAY
@@ -493,7 +492,7 @@ static __inline__ void snd_msnd_eval_dsp_msg(register WORD wMessage)
 }
 
 //static int InInterrupt = 0;
-irqreturn_t snd_msnd_interrupt(int irq, void *dev_id, struct pt_regs *regs)
+irqreturn_t snd_msnd_interrupt(int irq, void *dev_id)
 {
 	/*if( InInterrupt){
 		printk(  "INTERRUPT in InInterrupt\n");
@@ -702,37 +701,48 @@ static int __init snd_msnd_calibrate_adc( WORD srate)
 
 static int upload_dsp_code(void)
 {
+#ifndef HAVE_DSPCODEH
+	const struct firmware *init_fw, *perm_fw;
+#endif
+	int err;
+
 	outb(HPBLKSEL_0, dev.io + HP_BLKS);
 #ifndef HAVE_DSPCODEH
-	INITCODESIZE = mod_firmware_load(INITCODEFILE, &INITCODE);
-	if (!INITCODE) {
+	err = request_firmware(&init_fw, INITCODEFILE, dev.card->dev);
+	if (err < 0) {
 		printk(KERN_ERR LOGNAME ": Error loading " INITCODEFILE);
-		return -EBUSY;
+		goto cleanup1;
 	}
-
-	PERMCODESIZE = mod_firmware_load(PERMCODEFILE, &PERMCODE);
-	if (!PERMCODE) {
+	err = request_firmware(&perm_fw, PERMCODEFILE, dev.card->dev);
+	if (err < 0) {
 		printk(KERN_ERR LOGNAME ": Error loading " PERMCODEFILE);
-		vfree_nocheck(INITCODE);
-		return -EBUSY;
+		goto cleanup;
 	}
+	INITCODE = init_fw->data;
+	INITCODESIZE = init_fw->size;
+	PERMCODE = perm_fw->data;
+	PERMCODESIZE = perm_fw->size;
 #endif
 	isa_memcpy_toio(dev.base, PERMCODE, PERMCODESIZE);
 	if (snd_msnd_upload_host(&dev, INITCODE, INITCODESIZE) < 0) {
 		printk(KERN_WARNING LOGNAME ": Error uploading to DSP\n");
-		return -ENODEV;
+		err = -ENODEV;
+		goto cleanup;
 	}
 #ifdef HAVE_DSPCODEH
 	printk(KERN_INFO LOGNAME ": DSP firmware uploaded (resident)\n");
 #else
 	printk(KERN_INFO LOGNAME ": DSP firmware uploaded\n");
 #endif
+	err = 0;
 
+cleanup:
 #ifndef HAVE_DSPCODEH
-	vfree_nocheck(INITCODE);
-	vfree_nocheck(PERMCODE);
+	release_firmware(perm_fw);
+cleanup1:
+	release_firmware(init_fw);
 #endif
-	return 0;
+	return err;
 }
 
 #ifdef MSND_CLASSIC
@@ -801,12 +811,12 @@ static int snd_msnd_dsp_full_reset(void)
 	return rv;
 }
 
-static int snd_msnd_dev_free( snd_device_t *device){
+static int snd_msnd_dev_free( struct snd_device *device){
 	snd_printd( "snd_msnd_dev_free()\n");
 	return 0;
 }
 
-static snd_pcm_hardware_t snd_msnd_playback =
+static struct snd_pcm_hardware snd_msnd_playback =
 {
 	.info =			( SNDRV_PCM_INFO_MMAP | SNDRV_PCM_INFO_INTERLEAVED | SNDRV_PCM_INFO_MMAP_VALID),
 	.formats =		( SNDRV_PCM_FMTBIT_U8 | SNDRV_PCM_FMTBIT_S16_LE),
@@ -823,7 +833,7 @@ static snd_pcm_hardware_t snd_msnd_playback =
 	.fifo_size =		0,
 };
 
-static snd_pcm_hardware_t snd_msnd_capture =
+static struct snd_pcm_hardware snd_msnd_capture =
 {
 	.info =			(SNDRV_PCM_INFO_MMAP | SNDRV_PCM_INFO_INTERLEAVED |
 				 SNDRV_PCM_INFO_MMAP_VALID),
@@ -847,7 +857,7 @@ static unsigned int rates[7] = {
 	32000, 44100, 48000
 };
 
-static snd_pcm_hw_constraint_list_t hw_constraints_rates = {
+static struct snd_pcm_hw_constraint_list hw_constraints_rates = {
 	.count = 7,
 	.list = rates,
 	.mask = 0,
@@ -875,7 +885,7 @@ static void snd_msnd_pcm_timer_function(unsigned long data)
 	spin_unlock_irq( &dpcm->lock);	*/
 }
 
-static void snd_msnd_runtime_free( snd_pcm_runtime_t *runtime)
+static void snd_msnd_runtime_free( struct snd_pcm_runtime *runtime)
 {
 	snd_msndpinnacle_pcm_t* dpcm = runtime->private_data;
 	kfree( dpcm);
@@ -921,9 +931,9 @@ static void snd_msnd_generate_sine(signed short *samples, int channels,
 #endif
 
 
-static int snd_msnd_playback_open( snd_pcm_substream_t *substream)
+static int snd_msnd_playback_open( struct snd_pcm_substream *substream)
 {
-	snd_pcm_runtime_t*	runtime = substream->runtime;
+	struct snd_pcm_runtime *	runtime = substream->runtime;
 	snd_msndpinnacle_pcm_t*	dpcm;
 
 	//snd_printd( "snd_msnd_playback_open()\n");
@@ -953,7 +963,7 @@ static int snd_msnd_playback_open( snd_pcm_substream_t *substream)
 	return 0;
 }
 
-static int snd_msnd_playback_close( snd_pcm_substream_t *substream)
+static int snd_msnd_playback_close( struct snd_pcm_substream *substream)
 {
 	//snd_printd( "snd_msnd_playback_close()\n");
 	snd_msnd_disable_irq( &dev);
@@ -962,7 +972,7 @@ static int snd_msnd_playback_close( snd_pcm_substream_t *substream)
 }
 
 
-static int snd_msnd_playback_hw_params( snd_pcm_substream_t * substream, snd_pcm_hw_params_t * params)
+static int snd_msnd_playback_hw_params( struct snd_pcm_substream *substream, struct snd_pcm_hw_params *params)
 {
 	/*	snd_printd( " dpcm = %X M=%X\n", (unsigned)substream->runtime->private_data,
 			*((unsigned*)substream->runtime->private_data-1));
@@ -1006,15 +1016,15 @@ static int snd_msnd_playback_hw_params( snd_pcm_substream_t * substream, snd_pcm
 
 
 
-/*static int snd_msnd_playback_hw_free( snd_pcm_substream_t *substream)
+/*static int snd_msnd_playback_hw_free( struct snd_pcm_substream *substream)
 {
 	snd_printd( "snd_msnd_playback_hw_free()\n");
 	return 0;
 } */
 
-static int snd_msnd_playback_prepare( snd_pcm_substream_t * substream)
+static int snd_msnd_playback_prepare( struct snd_pcm_substream *substream)
 {
-	snd_pcm_runtime_t *runtime = substream->runtime;
+	struct snd_pcm_runtime *runtime = substream->runtime;
 	snd_msndpinnacle_pcm_t*	dpcm = runtime->private_data;
 	//snd_printd( "snd_msnd_playback_prepare()\n");
 
@@ -1028,7 +1038,7 @@ static int snd_msnd_playback_prepare( snd_pcm_substream_t * substream)
 }
 
 //static int played_bytes;
-static int snd_msnd_playback_trigger( snd_pcm_substream_t * substream, int cmd)
+static int snd_msnd_playback_trigger( struct snd_pcm_substream *substream, int cmd)
 {
 	int	result = 0;
 	//spin_lock(&chip->reg_lock);
@@ -1048,7 +1058,7 @@ static int snd_msnd_playback_trigger( snd_pcm_substream_t * substream, int cmd)
 		clear_bit( F_WRITING, &dev.flags);
 		snd_msnd_send_dsp_cmd(&dev, HDEX_PLAY_STOP);
 	} else {
-		snd_printd( "snd_msnd_playback_trigger( \?\?\?\?\?)\n");
+		snd_printd( "snd_msnd_playback_trigger( ?????)\n");
 		result = -EINVAL;
 	}
 	InTrigger = 0;             // interrupt diagnostic, comment this out later
@@ -1058,9 +1068,9 @@ static int snd_msnd_playback_trigger( snd_pcm_substream_t * substream, int cmd)
 	return result;
 }
 
-static snd_pcm_uframes_t snd_msnd_playback_pointer( snd_pcm_substream_t * substream)
+static snd_pcm_uframes_t snd_msnd_playback_pointer( struct snd_pcm_substream *substream)
 {
-	snd_pcm_runtime_t*	runtime =	substream->runtime;
+	struct snd_pcm_runtime *	runtime =	substream->runtime;
 	//snd_printd( "snd_msnd_playback_pointer()\n");
 	//snd_msndpinnacle_pcm_t*	dpcm =		runtime->private_data;
 
@@ -1115,7 +1125,7 @@ static snd_pcm_uframes_t snd_msnd_playback_pointer( snd_pcm_substream_t * substr
 }
 
 
-static snd_pcm_ops_t snd_msnd_playback_ops = {
+static struct snd_pcm_ops snd_msnd_playback_ops = {
 	.open =		snd_msnd_playback_open,
 	.close =	snd_msnd_playback_close,
 	.ioctl =	snd_pcm_lib_ioctl,
@@ -1126,9 +1136,9 @@ static snd_pcm_ops_t snd_msnd_playback_ops = {
 	.pointer =	snd_msnd_playback_pointer,
 };
 
-static int snd_msnd_capture_open(snd_pcm_substream_t * substream)
+static int snd_msnd_capture_open(struct snd_pcm_substream *substream)
 {
-	snd_pcm_runtime_t*	runtime = substream->runtime;
+	struct snd_pcm_runtime *	runtime = substream->runtime;
 	snd_msndpinnacle_pcm_t*	dpcm;
 	//snd_printd( "snd_msnd_capture_open()\n");
 
@@ -1153,7 +1163,7 @@ static int snd_msnd_capture_open(snd_pcm_substream_t * substream)
 	return 0;
 }
 
-static int snd_msnd_capture_close( snd_pcm_substream_t * substream)
+static int snd_msnd_capture_close( struct snd_pcm_substream *substream)
 {
 	//snd_printd( "snd_msnd_capture_close()\n");
 	snd_msnd_disable_irq( &dev);
@@ -1161,9 +1171,9 @@ static int snd_msnd_capture_close( snd_pcm_substream_t * substream)
 	return 0;
 }
 
-static int snd_msnd_capture_prepare( snd_pcm_substream_t * substream)
+static int snd_msnd_capture_prepare( struct snd_pcm_substream *substream)
 {
-	snd_pcm_runtime_t *runtime = substream->runtime;
+	struct snd_pcm_runtime *runtime = substream->runtime;
 	snd_msndpinnacle_pcm_t*	dpcm = runtime->private_data;
 	//snd_printd( "snd_msnd_capture_prepare()\n");
 
@@ -1176,7 +1186,7 @@ static int snd_msnd_capture_prepare( snd_pcm_substream_t * substream)
 	return 0;
 }
 
-static int snd_msnd_capture_trigger( snd_pcm_substream_t * substream, int cmd)
+static int snd_msnd_capture_trigger( struct snd_pcm_substream *substream, int cmd)
 {
 	if( cmd == SNDRV_PCM_TRIGGER_START){
 		//snd_card_dummy_pcm_timer_start(substream);
@@ -1196,16 +1206,16 @@ static int snd_msnd_capture_trigger( snd_pcm_substream_t * substream, int cmd)
 }
 
 
-static snd_pcm_uframes_t snd_msnd_capture_pointer( snd_pcm_substream_t * substream)
+static snd_pcm_uframes_t snd_msnd_capture_pointer( struct snd_pcm_substream *substream)
 {
-	snd_pcm_runtime_t *runtime = substream->runtime;
+	struct snd_pcm_runtime *runtime = substream->runtime;
 //	snd_card_dummy_pcm_t *dpcm = runtime->private_data;
 
 	return bytes_to_frames( runtime, dev.captureDMAPos);
 }
 
 
-static int snd_msnd_capture_hw_params( snd_pcm_substream_t * substream, snd_pcm_hw_params_t * params)
+static int snd_msnd_capture_hw_params( struct snd_pcm_substream *substream, struct snd_pcm_hw_params *params)
 {
 	int		i;
 	LPDAQD		lpDAQ =		dev.base + DARQ_DATA_BUFF;
@@ -1229,7 +1239,7 @@ static int snd_msnd_capture_hw_params( snd_pcm_substream_t * substream, snd_pcm_
 }
 
 
-static snd_pcm_ops_t snd_msnd_capture_ops = {
+static struct snd_pcm_ops snd_msnd_capture_ops = {
 	.open =		snd_msnd_capture_open,
 	.close =	snd_msnd_capture_close,
 	.ioctl =	snd_pcm_lib_ioctl,
@@ -1241,9 +1251,9 @@ static snd_pcm_ops_t snd_msnd_capture_ops = {
 };
 
 
-int snd_msnd_pcm( multisound_dev_t *chip, int device, snd_pcm_t **rpcm)
+int snd_msnd_pcm( multisound_dev_t *chip, int device, struct snd_pcm **rpcm)
 {
-	snd_pcm_t*	pcm;
+	struct snd_pcm *	pcm;
 	int		err;
 
 	if( ( err = snd_pcm_new( chip->card, "MSNDPINNACLE", 0/*device*/, 1, 1, &pcm)) < 0)
@@ -1267,8 +1277,8 @@ int snd_msnd_pcm( multisound_dev_t *chip, int device, snd_pcm_t **rpcm)
 static int __init snd_msnd_attach(void)
 {
 	int err;
-	snd_card_t *card;
-	static snd_device_ops_t ops = {
+	struct snd_card *card;
+	static struct snd_device_ops ops = {
 		.dev_free =      snd_msnd_dev_free,
 		};
 
@@ -1279,7 +1289,7 @@ static int __init snd_msnd_attach(void)
 	}
 	request_region(dev.io, dev.numio, dev.name);
 
-	dev.mappedbase = (void*)( __ISA_IO_base + dev.base);
+	dev.mappedbase = (void __force *)( __ISA_IO_base + dev.base);
 
 	/* geht nich!!!!!
 		useless, don't do this at home.
@@ -1534,23 +1544,9 @@ static int __init snd_msnd_pinnacle_cfg_devices(int cfg, int reset, msnd_pinnacl
 MODULE_AUTHOR			("Karsten Wiese <annabellesgarden@yahoo.de>");
 MODULE_DESCRIPTION		("Turtle Beach " LONGNAME " Linux Driver");
 MODULE_LICENSE("GPL");
-
-
-MODULE_PARM			(io, "i");
-MODULE_PARM			(irq, "i");
-MODULE_PARM			(mem, "i");
-MODULE_PARM			(write_ndelay, "i");
-MODULE_PARM			(calibrate_signal, "i");
-#ifndef MSND_CLASSIC
-MODULE_PARM			(digital, "i");
-MODULE_PARM			(cfg, "i");
-MODULE_PARM			(reset, "i");
-MODULE_PARM			(mpu_io, "i");
-MODULE_PARM			(mpu_irq, "i");
-MODULE_PARM			(ide_io0, "i");
-MODULE_PARM			(ide_io1, "i");
-MODULE_PARM			(ide_irq, "i");
-MODULE_PARM			(joystick_io, "i");
+#ifndef HAVE_DSPCODEH
+MODULE_FIRMWARE(INITCODEFILE);
+MODULE_FIRMWARE(PERMCODEFILE);
 #endif
 
 static int io __initdata =		-1;
@@ -1579,6 +1575,24 @@ static int digital __initdata = 0;
 
 static int calibrate_signal __initdata = 0;
 
+module_param(io, int, S_IRUGO);
+module_param(irq, int, S_IRUGO);
+module_param(mem, int, S_IRUGO);
+module_param(write_ndelay, int, S_IRUGO);
+module_param(calibrate_signal, int, S_IRUGO);
+#ifndef MSND_CLASSIC
+module_param(digital, int, S_IRUGO);
+module_param(cfg, int, S_IRUGO);
+module_param(reset, int, S_IRUGO);
+module_param(mpu_io, int, S_IRUGO);
+module_param(mpu_irq, int, S_IRUGO);
+module_param(ide_io0, int, S_IRUGO);
+module_param(ide_io1, int, S_IRUGO);
+module_param(ide_irq, int, S_IRUGO);
+module_param(joystick_io, int, S_IRUGO);
+
+#endif
+
 #else /* not a module */
 
 static int write_ndelay __initdata =	-1;
