@@ -33,15 +33,11 @@ If USE_ZLIB is defined, hpizlib.c must also be linked
 
 //#include <stdio.h>
 #include "hpi.h"
-#include "hpios.h"
 
-#include <linux/version.h>
-#if ( LINUX_VERSION_CODE >= KERNEL_VERSION ( 2 , 5 , 0 ) )
-#include <linux/device.h>
+#ifndef DISABLE_PRAGMA_PACK1
+#pragma pack(push,1)
 #endif
 
-#include <linux/firmware.h>
-
 /** Descriptor for dspcode from firmware loader */
 struct DSP_CODE_FIRMWARE {
 	const struct firmware *psFirmware;	// Firmware descriptor
@@ -49,11 +45,16 @@ struct DSP_CODE_FIRMWARE {
 	long int dwBlockLength;	//!< Expected number of words in the whole dsp code,INCL header
 	long int dwWordCount;	//!< Number of words read so far
 	u32 dwVersion;		//<! Version read from dsp code file
+	u32 dwCrc;		//<! CRC read from dsp code file
 };
 
 /* DSP CODE IS LOADED FROM FILE DSPnnnn.BIN */
 typedef struct DSP_CODE_FIRMWARE DSP_CODE;
 
+#ifndef DISABLE_PRAGMA_PACK1
+#pragma pack(pop)
+#endif
+
 enum BootLoadFamily {
 	Load2200 = 0x2200,
 	Load4100 = 0x4100,
@@ -64,10 +65,12 @@ enum BootLoadFamily {
 	Load5000 = 0x5000,
 	Load6200 = 0x6200,
 	Load6413 = 0x6413,
+	Load6600 = 0x6600,
 	Load8600 = 0x8600,
 	Load6205 = 0x6205,
 	Load8713 = 0x8713,
-	Load8800 = 0x8800
+	Load8800 = 0x8800,
+	Load8900 = 0x8900
 };
 
 /*! Prepare *psDspCode to refer to the requuested adapter.
@@ -76,7 +79,8 @@ Searches the file, or selects the appropriate linked array
 \return 0 for success, or error code if requested code is not available
 */
 short HpiDspCode_Open(u32 nAdapter,	//!< Adapter family
-		      DSP_CODE * psDspCode	//!< Pointer to DSP code control structure
+		      DSP_CODE * psDspCode,	//!< Pointer to DSP code control structure
+		      u32 * pdwOsErrorCode	//!< Pointer to dword to receive OS specific error code
     );
 
 /*! Close the DSP code file */
