@@ -97,6 +97,7 @@ static char *hiddendir = "..";
 static char *kernel_deps[] = {
 	/* buses */
 	"ISA",
+	"ISA_DMA_API",
 	"ISAPNP",
 	"EISA",
 	"PCI",
@@ -118,26 +119,31 @@ static char *kernel_deps[] = {
 	"X86",
 	"MIPS",
 	"MIPS64",
+	"SUPERH",
+	"SUPERH64",
 	"IA32_EMULATION",
 	/* architecture specific */
 	"ARCH_SA1100",
 	"ARCH_PXA",
 	"X86_PC9800",
+	"SH_DREAMCAST",
 	/* other drivers */
 	"RTC",
 	"GAMEPORT",
 	"VIDEO_DEV",
+	"VIDEO_V4L1",
 	"FW_LOADER",
-	"ASIHPK", /* external tree */
 	/* some flags */
 	"EXPERIMENTAL",
+	/* sound common */
+	"AC97_BUS",
 	NULL
 };
 
 /* % -> always true */
 /* # -> define */
 static char *no_cards[] = {
-	"%SOUND",
+	"%#SOUND",
 	"SOUND_PRIME",
 	"%SND",
 	"#SND_TIMER",
@@ -147,23 +153,45 @@ static char *no_cards[] = {
 	"SND_SEQUENCER",
 	"SND_MIXER_OSS",
 	"SND_PCM_OSS",
+	"SND_PCM_OSS_PLUGINS",
 	"SND_SEQUENCER_OSS",
 	"SND_OSSEMUL",
 	"SND_RTCTIMER",
 	"SND_HPET",
+	"SND_DYNAMIC_MINORS",
 	"SND_DEBUG",
 	"SND_DEBUG_MEMORY",
 	"SND_DEBUG_DETECT",
+	"SND_VERBOSE_PROCFS",
 	"SND_VERBOSE_PRINTK",
 	"SND_BIT32_EMUL",
+	"#SND_SB_COMMON",
+	"#SND_SB8_DSP",
+	"#SND_SB16_DSP",
 	"#SND_OPL3_LIB",
 	"#SND_OPL4_LIB",
 	"#SND_VX_LIB",
 	"#SND_MPU401_UART",
 	"#SND_GUS_SYNTH",
+	"#AC97_BUS",
 	"#SND_AC97_CODEC",
+	"#SND_AT91_SOC_SSC",
+	"#SND_SOC_AC97_CODEC",
+	"#SND_SOC_WM8731",
+	"#SND_SOC_WM8750",
+	"#SND_SOC_WM8753",
+	"#SND_SOC_WM9712",
 	"#SND_PXA2XX_PCM",
-	"#SND_GENERIC_PM",
+	"#SND_PXA2XX_AC97",
+	"#SND_PXA2XX_SOC_AC97",
+	"#SND_PXA2XX_SOC_I2S",
+	"#SND_S3C24XX_SOC_I2S",
+	"#SND_S3C2443_SOC_AC97",
+	"#SND_S3C24XX_SOC_NEO1973_WM8753",
+	"#SND_S3C24XX_SOC_SMDK2443_WM9710",
+	"#SND_AT91_SOC_I2S",
+	"#SND_SOC_SH4_HAC",
+	"#SND_SOC_SH4_SSI",
 	NULL
 };
 
@@ -185,9 +213,11 @@ static int read_file(const char *filename, struct cond **template)
 	
 	fullfile = malloc(strlen(basedir) + strlen(hiddendir) + 1 + strlen(filename) + 1);
 	sprintf(fullfile, "%s/%s", basedir, filename);
-	if ((err = read_file_1(fullfile, template)) < 0) {
-		free(fullfile);
-		return err;
+	if (access(fullfile, R_OK) == 0) {
+		if ((err = read_file_1(fullfile, template)) < 0) {
+			free(fullfile);
+			return err;
+		}
 	}
 	if (!strncmp(filename, "core/", 5))
 		sprintf(fullfile, "%s/acore/%s", hiddendir, filename + 5);
@@ -301,17 +331,22 @@ static int read_file_1(const char *filename, struct cond **template)
 				dep->is_bool = 1;
 			else if (!strncmp(buffer, "\tdepends on ", 12))
 				add_dep(dep, buffer + 12, *template);
+			else if (!strncmp(buffer, "\tdepends ", 9))
+				add_dep(dep, buffer + 9, *template);
 			else if (!strncmp(buffer, "\tselect ", 8))
 				add_select(dep, buffer + 8, *template);
 			continue;
 		case READ_STATE_MENU:
-			if (!strncmp(buffer, "\tdepends on ", 12)) {
+			if (!strncmp(buffer, "\tdepends ", 9)) {
 				struct cond *ntemplate;
 				if (strcmp((*template)->name, "EMPTY")) {
 					fprintf(stderr, "Menu consistency error\n");
 					exit(EXIT_FAILURE);
 				}
-				ntemplate = create_cond(buffer + 12);
+				if (! strncmp(buffer + 9, "on ", 3))
+					ntemplate = create_cond(buffer + 12);
+				else
+					ntemplate = create_cond(buffer + 9);
 				free((*template)->name);
 				(*template)->name = ntemplate->name;
 				(*template)->dep = ntemplate->dep;
@@ -400,15 +435,21 @@ static struct cond * create_cond(char *line)
 			nomem();
 		if (strcmp(cond->name, "EMPTY"))
 			find_or_create_dep(word);
-		if (get_word(line, word)) {
+		while (get_word(line, word)) {
 			if (!strcmp(word, "&&"))
 				cond->type = COND_AND;
 			else if (!strcmp(word, "||"))
 				cond->type = COND_OR;
+			else if (!strcmp(word, "!=")) {
+				get_word(line, word);
+				continue;
+			} else if (!strcmp(word, "!=n"))
+				continue;
 			else {
 				fprintf(stderr, "Wrong condition %s\n", word);
 				exit(EXIT_FAILURE);
 			}
+			break;
 		}
 	}
 	free(word);
@@ -513,9 +554,28 @@ static struct cond *join_cond(struct cond *cond1, struct cond *cond2)
 static void add_dep(struct dep * dep, char *line, struct cond *template)
 {
 	template = duplicate_cond(template);
+	if (dep->cond)
+		template = join_cond(dep->cond, template);
 	dep->cond = join_cond(template, create_cond(line));
 }
 
+/* Is the selected item ALSA-specific? */
+static int is_alsa_item(const char *word)
+{
+	static const char *known_items[] = {
+		"AC97_BUS",
+		NULL
+	};
+	const char **p;
+
+	if (!strncmp(word, "SND_", 4))
+		return 1;
+	for (p = known_items; *p; p++)
+		if (!strcmp(word, *p))
+			return 1;
+	return 0;
+}
+
 // Add a new forced (selected) dependency to the current one
 static void add_select(struct dep * dep, char *line, struct cond *template)
 {
@@ -526,7 +586,7 @@ static void add_select(struct dep * dep, char *line, struct cond *template)
 	if (word == NULL)
 		nomem();
 	get_word(line, word);
-	if (strncmp(word, "SND_", 4)) {
+	if (!is_alsa_item(word)) {
 		add_dep(dep, word, template);
 		free(word);
 		return;
@@ -608,12 +668,13 @@ static struct dep *find_dep(char *parent, char *depname)
 	int idx;
 
 	while (temp_dep) {
-		// fprintf(stderr, "depname = '%s', name = '%s'\n", depname, temp_dep->name);
+		//fprintf(stderr, "depname = '%s', name = '%s'\n", depname, temp_dep->name);
 		if (!strcmp(depname, temp_dep->name))
 			return temp_dep;
 		temp_dep = temp_dep->next;
 	}
 	for (idx = 0; kernel_deps[idx]; idx++) {
+		//fprintf(stderr, "kdepname = '%s', name = '%s'\n", depname, kernel_deps[idx]);
 		if (!strcmp(kernel_deps[idx], depname))
 			return NULL;
 	}
@@ -882,10 +943,103 @@ static void sel_remove_hitflags(void)
 	}
 }
 
+struct kconfig_verdep {
+	const char *name;
+	const char *version;
+	struct kconfig_verdep *next;
+};
+
+static struct kconfig_verdep *version_deps;
+
+static char *get_token(char **pp, int need_space)
+{
+	char *token;
+	char *p = *pp;
+
+	for (; isspace(*p); p++)
+		if (!*p)
+			return NULL;
+	if (*p == '#' || *p == '%')
+		return NULL;
+	token = p;
+	for (; !isspace(*p); p++)
+		if (!*p) {
+			if (need_space)
+				return NULL;
+			*pp = p;
+			return token;
+		}
+	*p++ = 0;
+	*pp = p;
+	return token;
+}
+
+static int read_version_deps(const char *fname)
+{
+	FILE *fp;
+	char buf[128], *p, *name, *val;
+	struct kconfig_verdep *ver;
+
+	fp = fopen(fname, "r");
+	if (!fp) {
+		fprintf(stderr, "cannot open %s\n", fname);
+		return -ENODEV;
+	}
+	while (fgets(buf, sizeof(buf), fp)) {
+		p = buf;
+		name = get_token(&p, 1);
+		if (!name)
+			continue;
+		val = get_token(&p, 0);
+		if (!val)
+			continue;
+		ver = malloc(sizeof(*ver));
+		if (!ver)
+			return -ENOMEM;
+		ver->name = strdup(name);
+		ver->version = strdup(val);
+		ver->next = version_deps;
+		version_deps = ver;
+	}
+	fclose(fp);
+	return 0;
+}
+
+static const char *get_version_dep(const char *name)
+{
+	const struct kconfig_verdep *p;
+	for (p = version_deps; p; p = p->next)
+		if (! strcmp(name, p->name))
+			return p->version;
+	return NULL;
+}
+
+static void print_version_dep(const char *ver)
+{
+	const char *p;
+	printf("test \"$kversion.$kpatchlevel\" = \"");
+	for (p = ver; *p; p++) {
+		putchar(*p);
+		if (*p == '.') {
+			p++;
+			break;
+		}
+	}
+	for (; *p && *p != '.'; p++)
+		putchar(*p);
+	putchar('"');
+	if (*p) {
+		printf(" -a $ksublevel -ge ");
+		for (p++; *p && *p != '.'; p++)
+			putchar(*p);
+	}
+}
+
 static void sel_print_acinclude(struct sel *sel)
 {
 	struct dep * dep;
 	struct sel * nsel;
+	const char *ver;
 
 	dep = sel->dep;
 	if (dep == NULL)
@@ -899,7 +1053,13 @@ static void sel_print_acinclude(struct sel *sel)
 		sel_print_acinclude(nsel);
 		if (!nsel->dep->hitflag) {
 			nsel->dep->hitflag = 1;
-			printf("      CONFIG_%s=\"%c\"\n", nsel->name,
+			printf("      ");
+			ver = get_version_dep(nsel->name);
+			if (ver) {
+				print_version_dep(ver);
+				printf(" && ");
+			}
+			printf("CONFIG_%s=\"%c\"\n", nsel->name,
 				(nsel->dep && nsel->dep->is_bool) ? 'y' : 'm');
 		}
 	}
@@ -912,6 +1072,7 @@ static void output_acinclude(void)
 	char *text;
 	struct cond *cond, *cond_prev;
 	struct sel *sel;
+	const char *ver;
 	int j;
 	
 	printf("dnl ALSA soundcard configuration\n");
@@ -969,6 +1130,21 @@ static void output_acinclude(void)
 			put_if = 1;
 			cond_prev = cond;
 		}
+		text = convert_to_config_uppercase("", tempdep->name);
+		ver = get_version_dep(text);
+		if (ver) {
+			if (!put_if)
+				printf("    if ");
+			else {
+				printf(cond_prev->type == COND_AND ? " &&" : " ||");
+				printf("\n      ");
+			}
+			printf("( ");
+			print_version_dep(ver);
+			printf(" )");
+			put_if = 1;
+		}
+		free(text);
 		if (put_if)
 			printf("; then\n");
 
@@ -978,7 +1154,13 @@ static void output_acinclude(void)
 		for (sel = tempdep->sel; sel; sel = sel->next) {
 			if (is_always_true(sel->dep))
 				continue;
-			printf("      CONFIG_%s=\"%c\"\n", sel->name,
+			printf("      ");
+			ver = get_version_dep(sel->name);
+			if (ver) {
+				print_version_dep(ver);
+				printf(" && ");
+			}
+			printf("CONFIG_%s=\"%c\"\n", sel->name,
 			       (sel->dep && sel->dep->is_bool) ? 'y' : 'm');
 		}
 		text = convert_to_config_uppercase("CONFIG_", tempdep->name);
@@ -1046,7 +1228,13 @@ static void output_acinclude(void)
 		for (sel = tempdep->sel; sel; sel = sel->next) {
 			if (is_always_true(sel->dep))
 				continue;
-			printf("      CONFIG_%s=\"%c\"\n", sel->name,
+			printf("      ");
+			ver = get_version_dep(sel->name);
+			if (ver) {
+				print_version_dep(ver);
+				printf(" && ");
+			}
+			printf("CONFIG_%s=\"%c\"\n", sel->name,
 			       (sel->dep && sel->dep->is_bool) ? 'y' : 'm');
 		}
 		text = convert_to_config_uppercase("CONFIG_", tempdep->name);
@@ -1068,6 +1256,9 @@ static void output_acinclude(void)
 
 	printf("AC_MSG_RESULT($SELECTED_CARDS)\n\n");
 	printf("CONFIG_SND=\"m\"\n");
+	printf("])\n\n");
+
+	printf("AC_DEFUN([ALSA_TOPLEVEL_DEFINES], [\n");
 	for (tempdep = all_deps; tempdep; tempdep = tempdep->next) {
 		if (!output_ac_define(tempdep))
 			continue;
@@ -1248,6 +1439,12 @@ int main(int argc, char *argv[])
 			argidx += 2;
 			continue;
 		}
+		if (strcmp(argv[argidx], "--versiondep") == 0) {
+			if (read_version_deps(argv[argidx + 1]) < 0)
+				exit(EXIT_FAILURE);
+			argidx += 2;
+			continue;
+		}
 		break;
 	}
 	if (strcmp(argv[argidx], "--acinclude") == 0)
